#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2016-2017 Adrian Perez <aperez@igalia.com>
#
# Distributed under terms of the GPLv3 license.

from gi.repository import Gtk, GLib
from .util import cachedproperty, desktop_is
import enum


class Status(enum.Enum):
    DISCONNECTED = "disconnected"
    CONNECTED = "connected"
    BLINKING = "blinking"


class StatusIconImpl(object):
    def __init__(self, delegate):
        self.delegate = delegate

    def __del__(self):
        self.delegate = None

    def set_tooltip(self, text):
        raise NotImplementedError

    def set_status(self, status):
        raise NotImplementedError


class StatusIconImplSNI(StatusIconImpl):
    ICON_PIXBUF_SIZE = 64  # This seems to be a reasonable size for all DEs.

    def __init__(self, delegate, context_menu, app, failure_callback):
        super().__init__(delegate)

        import gi
        gi.require_version("StatusNotifier", "1.0")
        from gi.repository import StatusNotifier

        if hasattr(StatusNotifier.Icon, "ATTENTION_ICON"):
            self.SNI_ATTENTION_ICON = StatusNotifier.Icon.ATTENTION_ICON
            self.SNI_ACTIVE_ICON = StatusNotifier.Icon.ICON
        else:
            self.SNI_ATTENTION_ICON = StatusNotifier.Icon.STATUS_NOTIFIER_ATTENTION_ICON
            self.SNI_ACTIVE_ICON = StatusNotifier.Icon.STATUS_NOTIFIER_ICON
        self.SNI_ATTENTION = StatusNotifier.Status.NEEDS_ATTENTION
        self.SNI_ACTIVE = StatusNotifier.Status.ACTIVE

        theme = Gtk.IconTheme.get_default()
        self._offline_icon_pixbuf = theme.load_icon("org.perezdecastro.Revolt",
                                                    self.ICON_PIXBUF_SIZE,
                                                    Gtk.IconLookupFlags.FORCE_SVG |
                                                    Gtk.IconLookupFlags.FORCE_SYMBOLIC)
        self._attention_icon_pixbuf = theme.load_icon("org.perezdecastro.Revolt-status-blink",
                                                      self.ICON_PIXBUF_SIZE,
                                                      Gtk.IconLookupFlags.FORCE_SVG |
                                                      Gtk.IconLookupFlags.FORCE_SYMBOLIC)
        self._online_icon_pixbuf = theme.load_icon("org.perezdecastro.Revolt-status-online",
                                                   self.ICON_PIXBUF_SIZE,
                                                   Gtk.IconLookupFlags.FORCE_SVG |
                                                   Gtk.IconLookupFlags.FORCE_SYMBOLIC)

        self._failure_callback = failure_callback
        self._sni = StatusNotifier.Item.new_from_pixbuf(app.get_application_id(),
                                                        StatusNotifier.Category.COMMUNICATIONS,
                                                        self._offline_icon_pixbuf)
        if not self._sni.set_context_menu(context_menu):
            # TODO: No DbusMenu support built into StatusIcon, we need to handle the
            #  "context-menu" signal ourselves. For now, fallback to use GtkStatusIcon
            raise RuntimeError("StatusNotifier does not support DbusMenu, falling back to GtkStatusIcon")

        self._sni.connect("registration-failed", self.__on_registration_failed)
        self._sni.connect("activate", self.__on_activate)
        self._sni.set_from_pixbuf(self.SNI_ATTENTION_ICON, self._attention_icon_pixbuf)
        self._sni.set_title("Revolt")
        self._sni.set_status(self.SNI_ACTIVE)
        self._sni.set_item_is_menu(False)
        self._sni.freeze_tooltip()
        self._sni.set_tooltip_title("Revolt")
        self._sni.thaw_tooltip()
        self._sni.register()

    def set_status(self, status):
        if status is Status.BLINKING:
            self._sni.set_status(self.SNI_ATTENTION)
        else:
            self._sni.set_status(self.SNI_ACTIVE)
            if status is Status.CONNECTED:
                self._sni.set_from_pixbuf(self.SNI_ACTIVE_ICON, self._online_icon_pixbuf)
            elif status is Status.DISCONNECTED:
                self._sni.set_from_pixbuf(self.SNI_ACTIVE_ICON, self._offline_icon_pixbuf)
            else:
                assert False, "Unrechable"

    def __on_registration_failed(self, sni, error):
        assert sni == self._sni
        print("StatusNotifier registration failed, falling back to GtkStatusIcon")
        self._failure_callback(self)

    def __on_activate(self, sni, x, y):
        assert sni == self._sni
        self.delegate.on_icon_activate(self)

    def set_tooltip(self, text):
        self._sni.freeze_tooltip()
        self._sni.set_tooltip_body("" if text is None else text)
        self._sni.thaw_tooltip()


class StatusIconImplGSI(StatusIconImpl):
    ICON_STATUS_NAMES = {
        Status.DISCONNECTED.value: "",
        Status.CONNECTED.value: "-status-online",
        "flip": "-status-blink",
        "flop": ""
    }

    def __init__(self, delegate, context_menu, app):
        super().__init__(delegate)
        self._status = Status.DISCONNECTED
        self._contextmenu = context_menu
        self._size = 16
        self._flipflop = True
        self._blinkmilliseconds = 500
        self._icondata = {}
        self.__load_icons(self._size, app)
        self._icon = Gtk.StatusIcon()
        self._icon.set_visible(True)
        self._icon.set_property("has-tooltip", True)
        self._icon.set_property("title", "Revolt")
        self._icon.connect("activate", self.__on_activate)
        self._icon.connect("popup-menu", self.__on_popup_menu)
        self._icon.connect("size-changed", self.__on_icon_size_change)

    def set_tooltip(self, text):
        if text is None:
            self._icon.set_tooltip_text("Revolt")
        else:
            self._icon.set_tooltip_markup("<b>Revolt</b>\n{!s}".format(text))

    def set_status(self, status):
        if status is Status.BLINKING:
            # We only want one blink callback active at a time.
            if self._status is not Status.BLINKING:
                GLib.timeout_add(self._blinkmilliseconds, self.__blink)
        else:
            GLib.timeout_add(2 * self._blinkmilliseconds, self.__draw_icon, status)
        self._status = status

    def __load_icons(self, size, app=None):
        if app is None:
            app = Gtk.Application.get_default()
        self._size = size
        theme = Gtk.IconTheme.get_default()
        for status, icon_suffix in self.ICON_STATUS_NAMES.items():
            icon_name = "org.perezdecastro.Revolt" + icon_suffix
            self._icondata[status] = theme.load_icon(icon_name, int(size),
                                                     Gtk.IconLookupFlags.FORCE_SVG |
                                                     Gtk.IconLookupFlags.FORCE_SYMBOLIC)

    def __draw_icon(self, status=None):
        if status is None:
            status = self._status
        if status is Status.BLINKING:
            if self._flipflop:
                self._icon.set_from_pixbuf(self._icondata["flip"])
            else:
                self._icon.set_from_pixbuf(self._icondata["flop"])
        else:
            self._icon.set_from_pixbuf(self._icondata[status.value])
        return False

    def __on_activate(self, icon):
        assert icon == self._icon
        self.delegate.on_icon_activate(self)

    def __on_popup_menu(self, icon, button, time):
        assert icon == self._icon
        self._contextmenu.show_all()
        self._contextmenu.popup(None, None, None, self._icon, button, time)

    def __on_icon_size_change(self, statusicon, size):
        if size > 31:
            icon_size = "32"
        elif size > 23:
            icon_size = "24"
        else:
            icon_size = "16"
        if desktop_is("kde"):  # KDE: see gajim bug #5476
            icon_size = "32"
        if desktop_is("mate"):
            icon_size = "16"
        self.__load_icons(icon_size)
        self.__draw_icon()

    def __blink(self):
        self._flipflop = not self._flipflop
        self.__draw_icon()
        return self._status is Status.BLINKING


class StatusIcon(object):
    def __init__(self, app, initial_status=Status.DISCONNECTED):
        self.status = Status(initial_status)
        self.__app = app
        self.__tooltip = None
        # Try using StatusNotifier first
        self._contextmenu.insert_action_group("app", app)
        try:
            self._impl = StatusIconImplSNI(self, self._contextmenu, app, self.__sni_failed)
        except Exception as e:
            print("StatusNotifier failed, using GtkStatusIcon instead -", str(e))
            self.__sni_failed(None)
        self.__configure_impl()

    def __sni_failed(self, sni_impl):
        # Use the (deprecated) GtkStatusIcon as fallback
        self._impl = StatusIconImplGSI(self, self._contextmenu, self.__app)
        self.__configure_impl()

    def __configure_impl(self):
        self._impl.set_tooltip(None)
        self.clear_notifications()

    @cachedproperty
    def _contextmenu(self):
        model = self.__app.get_menu_by_id("app-menu")
        if model is None:
            # If showing the application menu in the GNOME Shell top bar is
            # disabled, then GtkApplication won't load gtk/menus.ui
            # automatically, but we still need it for the context menu.
            (model,) = self.__app._build("gtk/menus.ui", "app-menu")
        return Gtk.Menu.new_from_model(model)

    def __add_notification_tooltip_text(self, text):
        if self.__tooltip is None:
            self.__tooltip = text
        else:
            self.__tooltip += "\n"
            self.__tooltip += text
        self._impl.set_tooltip(self.__tooltip)

    def __clear_notification_tooltip_text(self):
        self._impl.set_tooltip(None)
        self.__tooltip = None

    def set_status(self, status):
        status = Status(status)
        if status is not self.status:
            self.status = status
            self._impl.set_status(self.status)

    def add_notification(self, text):
        self.__add_notification_tooltip_text(text)
        self.set_status(Status.BLINKING)

    def clear_notifications(self):
        self.__clear_notification_tooltip_text()
        self.set_status(Status.CONNECTED)

    # Delegate methods.
    def on_icon_activate(self, icon_impl):
        self.clear_notifications()
        if self.__app.is_visible_and_focused():
            self.__app.hide()
        else:
            self.__app.show()



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2016 Adrian Perez <aperez@igalia.com>
#
# Distributed under terms of the GPLv3 license.

import gi
gi.require_versions(dict(WebKit2="4.0",
                         Gtk="3.0",
                         GLib="2.0"))


def main(program_path):
    # Honor CTRL+C http://stackoverflow.com/q/16410852
    import signal
    import sys
    from .app import RevoltApp
    signal.signal(signal.SIGINT, signal.SIG_DFL)
    sys.exit(RevoltApp(program_path).run(sys.argv[1:]))



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2017 Adrian Perez <aperez@igalia.com>
#
# Distributed under terms of the GPLv3 license.

from gi.repository import Gtk, Gdk, GLib


def __window_close(accel_group, window, key, modifiers):
    assert isinstance(window, Gtk.Window)
    window.close()


window_close_on_escape = Gtk.AccelGroup()
window_close_on_escape.connect(Gdk.KEY_Escape, 0, 0, __window_close)


def __window_modify_zoom(accel_group, window, key, modifiers):
    if key == Gdk.KEY_0:
        window.application.settings.reset("zoom-factor")
    else:
        zoom = window.application.settings.get_double("zoom-factor")
        if key == Gdk.KEY_plus:
            zoom += 0.1
        elif key == Gdk.KEY_minus:
            zoom -= 0.1
        else:
            assert False, "Unreachable"
        schema = window.application.settings.get_property("settings-schema")
        value = GLib.Variant.new_double(zoom)
        if schema.get_key("zoom-factor").range_check(value):
            window.application.settings.set_value("zoom-factor", value)


def __window_webview_reload(accel_group, window, key, modifiers):
    window.reload_element(bypass_cache=True)


window_keys = Gtk.AccelGroup()
window_keys.connect(Gdk.KEY_r, Gdk.ModifierType.CONTROL_MASK | Gdk.ModifierType.SHIFT_MASK, 0,
                    __window_webview_reload)
for key in (Gdk.KEY_plus, Gdk.KEY_minus, Gdk.KEY_0):
    window_keys.connect(key, Gdk.ModifierType.CONTROL_MASK, 0, __window_modify_zoom)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2016 Adrian Perez <aperez@igalia.com>
#
# Distributed under terms of the GPv3 license.

from gi.repository import Gdk
from gi.repository import Gtk
import os


if hasattr(Gtk, "show_uri_on_window"):
    def show_uri(parent, uri, timestamp=None):
        if timestamp is None:
            timestamp = Gdk.CURRENT_TIME
        Gtk.show_uri_on_window(parent, uri, timestamp)
else:
    def show_uri(parent, uri, timestamp=None):
        if timestamp is None:
            timestamp = Gdk.CURRENT_TIME
        Gtk.show_uri(None, uri, timestamp)


class CachedProperty(object):
    __slots__ = ("value", "get_value")

    INVALID = object()

    def __init__(self, f):
        self.value = self.INVALID
        self.get_value = f

    def __call__(self, obj):
        if self.value is self.INVALID:
            self.value = self.get_value(obj)
        return self.value


def cachedproperty(f, doc=None):
    return property(CachedProperty(f), doc=doc)


def desktop_is(desktopname):
    desktopname = desktopname.lower()
    if desktopname == "kde" and os.environ.get("KDE_FULL_SESSION") == "true":
        return True
    if desktopname == "mate" and os.environ.get("MATE_DESKTOP_SESSION_ID"):
        return True
    for desktopvarname in ["XDG_SESSION_DESKTOP", "DESKTOP_SESSION", "XDG_CURRENT_DESKTOP",
                           "XDG_SESSION_DESKTOP", "XDG_MENU_PREFIX", "GDMSESSION", "XDG_DATA_DIRS"]:
        desktopvarvalue = os.environ.get(desktopvarname)
        if desktopvarvalue and desktopname in desktopvarvalue.lower():
            return True
    return False



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2016-2017 Adrian Perez <aperez@igalia.com>
#
# Distributed under terms of the GPLv3 license.

from gi.repository import GLib, Gtk, Gio, WebKit2, GObject
from .util import cachedproperty, show_uri, desktop_is
from . import accelerators
from . import statusicon


class MainWindow(Gtk.ApplicationWindow):
    network_busy = GObject.Property(type=bool, default=False)

    def __init__(self, application, saved_state):
        self.application = application
        self.saved_state = saved_state
        Gtk.ApplicationWindow.__init__(self,
                                       application=application,
                                       icon_name="revolt",
                                       role="main-window",
                                       default_width=saved_state.get_uint("width"),
                                       default_height=saved_state.get_uint("height"))
        if self.saved_state.get_boolean("maximized"):
            self.maximize()
        self.saved_state.bind("maximized", self, "is-maximized", Gio.SettingsBindFlags.SET)

        if application.settings.get_boolean("use-header-bar"):
            self.set_titlebar(self.__make_headerbar())

        if application.settings.get_boolean("hide-on-window-close"):
            self.connect("delete-event", self.__hide_on_destroy)

        self.set_title(u"Revolt")
        application.add_window(self)
        self._webview = WebKit2.WebView(user_content_manager=self._user_content_manager,
                                        web_context=self._web_context)
        self._webview.connect("decide-policy", self.__on_decide_policy)
        self._webview.connect("context-menu", self.__on_context_menu)
        application.settings.bind("zoom-factor", self._webview, "zoom-level",
                                  Gio.SettingsBindFlags.GET)
        if hasattr(self._webview, "set_maintains_back_forward_list"):
            self._webview.set_maintains_back_forward_list(False)
        websettings = self._webview.get_settings()
        application.settings.bind("enable-developer-tools", websettings,
                                  "enable-developer-extras",
                                  Gio.SettingsBindFlags.GET)
        application.settings.bind("enable-developer-tools", websettings,
                                  "enable-write-console-messages-to-stdout",
                                  Gio.SettingsBindFlags.GET)

        self.add_accel_group(accelerators.window_keys)

        websettings.set_allow_file_access_from_file_urls(True)
        websettings.set_allow_modal_dialogs(False)  # TODO
        websettings.set_enable_fullscreen(False)
        websettings.set_enable_java(False)
        websettings.set_enable_media_stream(True)
        websettings.set_enable_page_cache(False)  # Single-page app
        websettings.set_enable_plugins(False)
        websettings.set_enable_smooth_scrolling(True)
        websettings.set_enable_webaudio(True)
        websettings.set_javascript_can_access_clipboard(True)
        websettings.set_minimum_font_size(12)  # TODO: Make it a setting
        websettings.set_property("enable-mediasource", True)

        if hasattr(websettings, "set_hardware_acceleration_policy"):
            websettings.set_hardware_acceleration_policy(WebKit2.HardwareAccelerationPolicy.ALWAYS)

        self._webview.show_all()
        self.add(self._webview)
        self.__connect_widgets()
        self.__notification_ids = set()

    def do_configure_event(self, event):
        result = Gtk.ApplicationWindow.do_configure_event(self, event)
        width, height = self.get_size()
        self.saved_state.set_uint("width", width)
        self.saved_state.set_uint("height", height)
        return result

    def __make_headerbar(self):
        header = Gtk.HeaderBar()
        header.set_show_close_button(True)
        header.get_style_context().add_class("revolt-slim")
        spinner = Gtk.Spinner()
        header.pack_end(spinner)
        self.bind_property("network-busy", spinner, "active",
                           GObject.BindingFlags.DEFAULT)
        header.show_all()
        return header

    @cachedproperty
    def _website_data_manager(self):
        from os import path as P
        print("Creating WebsiteDataManager...")
        app_id = self.application.get_application_id()
        cache_dir = P.join(GLib.get_user_cache_dir(), "revolt", app_id)
        data_dir = P.join(GLib.get_user_data_dir(), "revolt", app_id)
        return WebKit2.WebsiteDataManager(base_cache_directory=cache_dir,
                                          base_data_directory=data_dir)

    @cachedproperty
    def _web_context(self):
        print("Creating WebContext...")
        ctx = WebKit2.WebContext(website_data_manager=self._website_data_manager)
        ctx.set_spell_checking_enabled(False)
        ctx.set_tls_errors_policy(WebKit2.TLSErrorsPolicy.FAIL)
        return ctx

    @cachedproperty
    def _user_content_manager(self):
        mgr = WebKit2.UserContentManager()
        script = WebKit2.UserScript("Notification.requestPermission();",
                                    WebKit2.UserContentInjectedFrames.TOP_FRAME,
                                    WebKit2.UserScriptInjectionTime.START,
                                    None, None)
        mgr.add_script(script)
        return mgr

    def __on_decide_policy(self, webview, decision, decision_type):
        if decision_type == WebKit2.PolicyDecisionType.NAVIGATION_ACTION:
            if decision.get_navigation_type() == WebKit2.NavigationType.LINK_CLICKED:
                uri = decision.get_request().get_uri()
                if not uri.startswith(self.application.element_url):
                    show_uri(self, uri)
                    return True
        elif decision_type == WebKit2.PolicyDecisionType.NEW_WINDOW_ACTION:
            if decision.get_navigation_type() == WebKit2.NavigationType.LINK_CLICKED:
                show_uri(self, decision.get_request().get_uri())
                return True
        return False

    @cachedproperty
    def _context_menu_actions(self):
        action_list = []
        action = Gio.SimpleAction.new("preferences")
        action.connect("activate", self.application.on_app_preferences)
        action_list.append((action, "_Preferences"))
        action = Gio.SimpleAction.new("element-settings")
        action.connect("activate", self.application.on_element_settings)
        action_list.append((action, "_Element Settings"))
        return tuple(action_list)

    def __on_context_menu(self, webview, menu, event, hit_test):
        # Tweak built-in entries.
        for action in (WebKit2.ContextMenuAction.GO_BACK,
                       WebKit2.ContextMenuAction.GO_FORWARD,
                       WebKit2.ContextMenuAction.STOP):
            for index in range(menu.get_n_items()):
                item = menu.get_item_at_position(index)
                if action == item.get_stock_action():
                    menu.remove(item)
                    break
        # Add a separator.
        menu.append(WebKit2.ContextMenuItem.new_separator())
        # Append application-specfic entries.
        for (action, label) in self._context_menu_actions:
            menu.append(WebKit2.ContextMenuItem.new_from_gaction(action, label))
        return False

    def __on_has_toplevel_focus_changed(self, window, has_focus):
        assert window == self
        if window.has_toplevel_focus():
            # Clear the window's urgency hint
            window.set_urgency_hint(False)
            # Dismiss notifications
            for notification_id in self.__notification_ids:
                self.application.withdraw_notification(notification_id)
            self.__notification_ids.clear()
            self.application.statusicon.clear_notifications()

    def __on_load_changed(self, webview, event):
        if event == WebKit2.LoadEvent.FINISHED:
            self.network_busy = False
            self.application.statusicon.set_status(statusicon.Status.CONNECTED)
        else:
            self.network_busy = True
            self.application.statusicon.set_status(statusicon.Status.DISCONNECTED)

    @cachedproperty
    def _notification_icon(self):
        icon_id = self.application.get_application_id() + "-symbolic"
        return Gio.ThemedIcon.new(icon_id)

    def __on_show_notification(self, webview, notification):
        # TODO: Handle notification clicked, and so
        if not self.has_toplevel_focus():
            self.set_urgency_hint(True)
            notif = Gio.Notification.new(notification.get_title())
            notif.set_body(notification.get_body())
            # TODO: Use the avatar of the contact, if available.
            notif.set_icon(self._notification_icon)
            if not desktop_is("xfce"):  # Workaround for XFCE bug #13586
                notif.set_priority(Gio.NotificationPriority.HIGH)
            # use title as notification id:
            # allows to reuse one notification for the same conversation
            notification_id = notification.get_title()
            self.__notification_ids.add(notification_id)
            self.application.send_notification(notification_id, notif)
            self.application.statusicon.add_notification("%s: %s" % (notification.get_title(),
                                                                     notification.get_body()))
        return True

    def __on_permission_request(self, webview, request):
        if isinstance(request, WebKit2.NotificationPermissionRequest):
            request.allow()
            return True

    def __connect_widgets(self):
        self.connect("notify::has-toplevel-focus", self.__on_has_toplevel_focus_changed)
        self._webview.connect("load-changed", self.__on_load_changed)
        self._webview.connect("show-notification", self.__on_show_notification)
        self._webview.connect("permission-request", self.__on_permission_request)

    def __hide_on_destroy(self, widget, event):
        self.application.hide()
        return True

    def reload_element(self, bypass_cache=False):
        if bypass_cache:
            self._webview.reload_bypass_cache()
        else:
            self._webview.reload()

    def load_element(self):
        self._webview.load_uri(self.application.element_url)
        return self

    def load_settings_page(self):
        from urllib.parse import urlsplit, urlunsplit
        url = list(urlsplit(self._webview.get_uri()))
        url[-1] = "#settings"
        self._webview.load_uri(urlunsplit(url))

    def finish(self):
        self._webview.stop_loading()
        self.hide()
        self.destroy()
        del self._webview
        return self



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2016-2017 Adrian Perez <aperez@igalia.com>
#
# Distributed under terms of the GPLv3 license.

from os import environ
from gi.repository import Gtk, Gio, Gdk
from .statusicon import StatusIcon
from .window import MainWindow
from . import accelerators

DEFAULT_APP_ID = "org.perezdecastro.Revolt"
APP_ID = environ.get("REVOLT_OVERRIDE_APPLICATION_ID", DEFAULT_APP_ID).strip()

APP_COMMENTS = u"Desktop application for Element.io"
APP_WEBSITE = u"https://github.com/aperezdc/revolt"
APP_AUTHORS = (u"Adrián Pérez de Castro <aperez@igalia.com>",
               u"Jacobo Aragunde Pérez <jaragunde@igalia.com>",
               u"Carlos López Pérez <clopez@igalia.com>")


def _find_resources_path(program_path):
    from os import path as P
    devel = environ.get("__REVOLT_DEVELOPMENT")
    if devel and devel.strip():
        # Use the directory where the executable is located, most likely
        # a checkout of the Git repository.
        path = P.dirname(P.dirname(program_path))
    else:
        # Use an installed location: binary is in <prefix>/bin/revolt,
        # and resources in <prefix>/share/revolt/*
        path = P.join(P.dirname(P.dirname(program_path)), "share", "revolt")
    return P.abspath(P.join(path, DEFAULT_APP_ID + ".gresource"))


class RevoltApp(Gtk.Application):
    def __init__(self, program_path):
        Gio.Resource.load(_find_resources_path(program_path))._register()
        Gtk.Application.__init__(self, application_id=APP_ID,
                                 resource_base_path="/" + DEFAULT_APP_ID.replace(".", "/"),
                                 flags=Gio.ApplicationFlags.FLAGS_NONE)
        self.settings = Gio.Settings(schema_id=DEFAULT_APP_ID,
                                     path="/" + APP_ID.replace(".", "/") + "/")
        self.element_url = self.settings.get_string("riot-url")
        self.window = None
        self._last_window_geometry = None
        self.statusicon = None
        self.connect("shutdown", self.__on_shutdown)
        self.connect("activate", self.__on_activate)
        self.connect("startup", self.__on_startup)

    def __action(self, name, callback):
        action = Gio.SimpleAction.new(name)
        action.connect("activate", callback)
        self.add_action(action)

    def __on_startup(self, app):
        gtk_settings = Gtk.Settings.get_default()
        gtk_settings.set_property("gtk-dialogs-use-header",
                                  self.settings.get_boolean("use-header-bar"))
        css_provider = Gtk.CssProvider()
        css_provider.load_from_resource(self.get_resource_base_path() + "/gtk/custom.css")
        Gtk.StyleContext.add_provider_for_screen(Gdk.Screen.get_default(), css_provider,
                                                 Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION)
        self.statusicon = StatusIcon(self)
        self.__action("quit", lambda *arg: self.quit())
        self.__action("about", self.__on_app_about)
        self.__action("preferences", self.on_app_preferences)
        self.__action("element-settings", self.on_element_settings)

    def __on_shutdown(self, app):
        if self.window is not None:
            self.window.finish()

    def __on_activate(self, app):
        if self.window is None:
            saved_state_path = self.settings.get_property("path")
            saved_state_path += "saved-state/main-window/"
            saved_state = Gio.Settings(schema_id=DEFAULT_APP_ID + ".WindowState",
                                       path=saved_state_path)
            self.window = MainWindow(self, saved_state).load_element()
        self.show()

    def __on_app_about(self, action, param):
        dialog = Gtk.AboutDialog(transient_for=self.window,
                                 program_name=u"Revolt",
                                 authors=APP_AUTHORS,
                                 logo_icon_name="revolt-about",
                                 license_type=Gtk.License.GPL_3_0,
                                 comments=APP_COMMENTS,
                                 website=APP_WEBSITE)
        dialog.connect("response", lambda d, r: d.destroy())
        dialog.present()

    def _build(self, resource, *names):
        builder = Gtk.Builder.new_from_resource(self.get_resource_base_path() + "/" + resource)
        return (builder.get_object(name) for name in names)

    def on_app_preferences(self, action, param):
        window, url_entry, zoom_factor, zoom_factor_reset, devtools_toggle = \
                self._build("gtk/preferences.ui",
                            "settings-window",
                            "element-url-entry",
                            "zoom-factor",
                            "zoom-factor-reset",
                            "dev-tools-toggle")
        self.settings.bind("zoom-factor", zoom_factor, "value",
                           Gio.SettingsBindFlags.DEFAULT)
        self.settings.bind("enable-developer-tools", devtools_toggle, "active",
                           Gio.SettingsBindFlags.DEFAULT)
        zoom_factor_reset.connect("clicked", lambda button:
                                  self.settings.set_double("zoom-factor", 1.0))
        url_entry.set_text(self.element_url)

        def on_hide(window):
            new_url = url_entry.get_text()
            if new_url != self.element_url:
                self.settings.set_string("riot-url", new_url)
                self.element_url = new_url
                self.window.load_element()
        window.connect("hide", on_hide)
        window.add_accel_group(accelerators.window_close_on_escape)
        window.set_transient_for(self.window)
        window.present()

    def on_element_settings(self, action, param):
        self.show()
        self.window.load_settings_page()

    def __save_window_geometry(self):
        window_size = self.window.get_size()
        window_position = self.window.get_position()
        self._last_window_geometry = {"width": window_size.width,
                                      "height": window_size.height,
                                      "root_x": window_position.root_x,
                                      "root_y": window_position.root_y}

    def __restore_window_geometry(self):
        if not self._last_window_geometry:
            return
        self.window.resize(self._last_window_geometry["width"],
                           self._last_window_geometry["height"])
        self.window.move(self._last_window_geometry["root_x"],
                         self._last_window_geometry["root_y"])
        # invalidate _last_window_geometry after restoring to ensure to
        # not restore again if the users clicks on the status icon when
        # the window is not hidden (like when minimized, or not focused)
        # self.hide() will take care of setting a new saved geometry.
        self._last_window_geometry = None

    def show(self):
        self.__restore_window_geometry()
        self.window.show()
        self.window.present()

    def hide(self):
        self.__save_window_geometry()
        self.window.hide()

    def is_visible_and_focused(self):
        return self.window.props.visible and self.window.props.has_toplevel_focus
